// ATOM.H : the atom-related classes.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef ATOM_H
#define ATOM_H

class element;

class crec;
class atom;

/*################################################################################################*/

class bond;	// bond.h

#include "define.h"
#include "typedef.h"

#include <list>
#include <vector>
using namespace std;

/*################################################################################################*/

#define ELEMENT_SYMBOLS 110

/// An element class.

class element
{
	private:
	
	i32s atomic_number;			///< This is the atomic number, or NOT_DEFINED (equal to -1) for a dummy atom.
	
	static const char string[ELEMENT_SYMBOLS][32];
	static const char symbol[ELEMENT_SYMBOLS][4];
	
	static const fGL color[ELEMENT_SYMBOLS][3];
	
	static const fGL mass[ELEMENT_SYMBOLS];
	static const fGL vdwr[ELEMENT_SYMBOLS];

	public:
	
	static element current_element;		///< This is the current element that user has selected.
	
	public:
	
	element(void);
	element(char *);	///< By symbols.
	element(i32s);		///< By atomic numbers.
	~element(void);
	
	const i32s GetAtomicNumber(void) const;
	const char * GetString(void) const;
	const char * GetSymbol(void) const;	
	const fGL * GetColor(void) const;
	fGL GetAtomicMass(void) const;
	fGL GetVDWRadius(void) const;

// JUST DISABLED FOR A WHILE; remove the "const" settings from tables and write prefs directly there??? QUICKER!!!
// JUST DISABLED FOR A WHILE; remove the "const" settings from tables and write prefs directly there??? QUICKER!!!
// JUST DISABLED FOR A WHILE; remove the "const" settings from tables and write prefs directly there??? QUICKER!!!
// If you have a prefs object, we can check that for a user-defined radius or color
//	fGL GetVDWRadius(prefs *) const;
//	const fGL * GetColor(prefs *) const;
	
	void operator++(void);
	void operator--(void);
};

/*################################################################################################*/

/// A connectivity record class.
/** The connectivity records are maintained in the atom objects, to make quick browising of 
bond networks possible. */

class crec
{
//	protected:
	public:		// TODO : not properly divided in public/protected data.
	
	atom * atmr;
	bond * bndr;
	
	public:
	
	crec(void);
	crec(atom *, bond *);
	~crec(void);
	
//	bool operator<(const crec &) const { return false; }		// just for STL...
	bool operator==(const crec &) const;
};

/*################################################################################################*/

#define ATOMFLAG_SELECTED	0x00000001
// - locked???
// - protected???
// - submodel (qm/mm)???

#define ATOMFLAG_SF_HELIX	0x00010000
#define ATOMFLAG_SF_STRAND	0x00020000

/// An atom class.
/** Will store all data about an atom, including type, coordinates, and a connectivity 
record list for quick access to neighbor atoms. */

class atom
{
//	protected:
	public:		// TODO : not properly divided in public/protected data.
	
	element el;		// needed 2x for td???		// is saved into files!!!
	
	i32s atmtp;		// needed 2x for td???
	i32s sf_atmtp;						// is saved into files!!!
	
	f64 charge;		// needed 2x for td???		// is saved into files???
	f64 mass;		// needed 2x for td???
	
	f64 vdwr;		// only for graphics in aa, but for everything in sf!!!
	
	list<crec> cr_list;		///< Connectivity records.
	
protected:
	fGL * crd_table;		///< The actual coordinate data; see also model::PushCRDSets().
	i32u crd_table_size_loc;
public:

/// Molecule, chain, residue and atom ID-numbers.
/** model::GatherGroups() will maintain molecule numbers, and sequence builders will handle the rest. */
	i32s id[4];
	i32s res_id;		// 2x is NOT possible for this???
	
	i32s index;		///< Index of this entry in model::atom_list. Updated by model::UpdateIndex().
	
	i32u flags;		///< The flags that carry various information. These are saved into files.
	
	friend class model;	// model::PushCRDSets() needs this...
	
	public:
	
	atom(void);
	atom(element, const fGL *, i32u);
	atom(const atom &);
	~atom(void);
	
	const fGL * GetCRD(i32u);
	void SetCRD(i32s, fGL, fGL, fGL);
	
	bool operator<(const atom &) const;	///< Using id-numbers.
	bool operator==(const atom &) const;	///< Using pointers.
};

/*################################################################################################*/

typedef list<crec>::iterator iter_cl;		// cl = connectivity list
typedef list<atom>::iterator iter_al;		// al = atom list

/*################################################################################################*/

#endif	// ATOM_H

// eof
