// ENGINE.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "engine.h"

#include "model.h"

/*################################################################################################*/

setup::setup(model * p1, bool flag1, bool flag2)
{
	if (flag1 && flag2)	// detect if both are true...
	{
		cout << "BUG: both allqm and allmm requested at setup ctor." << endl;
		exit(EXIT_FAILURE);
	}
	
	mdl = p1;
	setup_all_qm = flag1;
	setup_all_mm = flag2;
	
	atmtab = NULL; natm = NOT_DEFINED;
	bndtab = NULL; nbnd = NOT_DEFINED;
	
	qm_atmtab = NULL; qm_natm = NOT_DEFINED;
	qm_bndtab = NULL; qm_nbnd = NOT_DEFINED;
	
	mm_atmtab = NULL; mm_natm = NOT_DEFINED;
	mm_bndtab = NULL; mm_nbnd = NOT_DEFINED;
	
	boundary_qmmm = NULL; qmmm_nbnd = NOT_DEFINED;
	
#ifdef ENABLE_SIMPLIFIED

	sf_atmtab = NULL; sf_natm = NOT_DEFINED;
	sf_bndtab = NULL; sf_nbnd = NOT_DEFINED;

	boundary_mmsf = NULL; mmsf_nbnd = NOT_DEFINED;

#endif	// ENABLE_SIMPLIFIED
	
	current_eng = NULL;
	current_eng_index = 0;
}

setup::~setup(void)
{
	if (atmtab != NULL) { delete[] atmtab; }
	if (bndtab != NULL) { delete[] bndtab; }
	
	if (qm_atmtab != NULL) { delete[] qm_atmtab; }
	if (qm_bndtab != NULL) { delete[] qm_bndtab; }
	
	if (mm_atmtab != NULL) { delete[] mm_atmtab; }
	if (mm_bndtab != NULL) { delete[] mm_bndtab; }
	
	if (boundary_qmmm != NULL) { delete[] boundary_qmmm; }
		
#ifdef ENABLE_SIMPLIFIED

	if (sf_atmtab != NULL) { delete[] sf_atmtab; }
	if (sf_bndtab != NULL) { delete[] sf_bndtab; }
	
	if (boundary_mmsf != NULL) { delete[] boundary_mmsf; }
	
#endif	// ENABLE_SIMPLIFIED
	
	DiscardCurrentEngine();
}

void setup::UpdateSetupInfo(void)
{
	if (atmtab != NULL) { delete[] atmtab; atmtab = NULL; }
	if (bndtab != NULL) { delete[] bndtab; bndtab = NULL; }

	natm = 0; nbnd = 0;
	
	if (qm_atmtab != NULL) { delete[] qm_atmtab; qm_atmtab = NULL; }
	if (qm_bndtab != NULL) { delete[] qm_bndtab; qm_bndtab = NULL; }
	
	if (mm_atmtab != NULL) { delete[] mm_atmtab; mm_atmtab = NULL; }
	if (mm_bndtab != NULL) { delete[] mm_bndtab; mm_bndtab = NULL; }
	
	if (boundary_qmmm != NULL) { delete[] boundary_qmmm; boundary_qmmm = NULL; }
	
	qm_natm = 0; qm_nbnd = 0;
	mm_natm = 0; mm_nbnd = 0;
	
	qmmm_nbnd = 0;
	
#ifdef ENABLE_SIMPLIFIED

	if (sf_atmtab != NULL) { delete[] sf_atmtab; sf_atmtab = NULL; }
	if (sf_bndtab != NULL) { delete[] sf_bndtab; sf_bndtab = NULL; }
	
	if (boundary_mmsf != NULL) { delete[] boundary_mmsf; boundary_mmsf = NULL; }
	
	sf_natm = 0; sf_nbnd = 0;
	
	mmsf_nbnd = 0;
	
#endif	// ENABLE_SIMPLIFIED
	
	// make the global tables...
	// make the global tables...
	// make the global tables...
	
	natm = mdl->GetAtomCount();
	nbnd = mdl->GetBondCount();
	
	atmtab = new (atom *)[natm];
	bndtab = new (bond *)[nbnd];
	
	i32u ac = 0; iter_al ita = mdl->GetAtomsBegin();
	while (ita != mdl->GetAtomsEnd())
	{
		if ((* ita).index != (i32s) ac)		// check the global index consistency...
		{
			cout << "BUG: mismatch in global indices at setup::UpdateSetupInfo()." << endl;
			exit(EXIT_FAILURE);
		}
		
		atmtab[ac] = & (* ita);
		ac++; ita++;
	}
	
	i32u bc = 0; iter_bl itb = mdl->GetBondsBegin();
	while (itb != mdl->GetBondsEnd())
	{
		bndtab[bc] = & (* itb);
		bc++; itb++;
	}
	
	// the setup_all_qm case...
	// the setup_all_qm case...
	// the setup_all_qm case...
	
	if (setup_all_qm)
	{
		qm_natm = mdl->GetAtomCount();
		qm_nbnd = mdl->GetBondCount();
		
		qm_atmtab = new (atom *)[qm_natm];
		qm_bndtab = new (bond *)[qm_nbnd];
		
		ac = 0; ita = mdl->GetAtomsBegin();
		while (ita != mdl->GetAtomsEnd())
		{
			qm_atmtab[ac++] = & (* ita);
			ita++;
		}
		
		bc = 0; itb = mdl->GetBondsBegin();
		while (itb != mdl->GetBondsEnd())
		{
			qm_bndtab[bc++] = & (* itb);
			itb++;
		}
		
		return;
	}
	
	// the setup_all_mm case...
	// the setup_all_mm case...
	// the setup_all_mm case...
	
	if (setup_all_mm)
	{
		mm_natm = mdl->GetAtomCount();
		mm_nbnd = mdl->GetBondCount();
		
		mm_atmtab = new (atom *)[mm_natm];
		mm_bndtab = new (bond *)[mm_nbnd];
		
		ac = 0; ita = mdl->GetAtomsBegin();
		while (ita != mdl->GetAtomsEnd())
		{
			mm_atmtab[ac++] = & (* ita);
			ita++;
		}
		
		bc = 0; itb = mdl->GetBondsBegin();
		while (itb != mdl->GetBondsEnd())
		{
			mm_bndtab[bc++] = & (* itb);
			itb++;
		}
		
		return;
	}
	
	// the default case...
	// the default case...
	// the default case...
	
// in this initial implementation,
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
//	1) consider the selected atoms to be a part of the QM model.
//	2) set those atoms, that have modified sf_atmtp, into sf category...

	// count the atoms and bonds.
	
	qm_natm = 0; qm_nbnd = 0;
	mm_natm = 0; mm_nbnd = 0;
	
#ifdef ENABLE_SIMPLIFIED

	sf_natm = 0; sf_nbnd = 0;

#endif	// ENABLE_SIMPLIFIED
	
	ita = mdl->GetAtomsBegin();
	while (ita != mdl->GetAtomsEnd())
	{
		bool ready = false;
		
#ifdef ENABLE_SIMPLIFIED
		
		if (!ready && (* ita).sf_atmtp != NOT_DEFINED)
		{
			sf_natm++;
			ready = true;
		}
		
#endif	// ENABLE_SIMPLIFIED

		if (!ready && ((* ita).flags & ATOMFLAG_SELECTED))
		{
			qm_natm++;
			ready = true;
		}
		else
		{
			mm_natm++;
			ready = true;
		}
		
		if (!ready) { cout << "BUG: a1-counting error." << endl; exit(EXIT_FAILURE); }
		ita++;
	}
	
	itb = mdl->GetBondsBegin();
	while (itb != mdl->GetBondsEnd())
	{
		bool ready = false;
		
#ifdef ENABLE_SIMPLIFIED
		
		bool sff1 = ((* itb).atmr[0]->sf_atmtp != NOT_DEFINED);
		bool sff2 = ((* itb).atmr[1]->sf_atmtp != NOT_DEFINED);
		if (!ready && sff1 && sff2)
		{
			sf_nbnd++;
			ready = true;
		}
		
#endif	// ENABLE_SIMPLIFIED
		
		bool qmf1 = ((* itb).atmr[0]->flags & ATOMFLAG_SELECTED);
		bool qmf2 = ((* itb).atmr[1]->flags & ATOMFLAG_SELECTED);
		if (!ready && qmf1 && qmf2)
		{
			qm_nbnd++;
			ready = true;
		}
		else
		{
			mm_nbnd++;		// not really correct yet!
			ready = true;
		}
		
		if (!ready) { cout << "BUG: b1-counting error." << endl; exit(EXIT_FAILURE); }
		itb++;
	}
	
//cout << "SETUP qma =" << qm_natm << " mma = " << mm_natm << endl; int zzz; cin >> zzz;
	
	// allocate the tables.
	
	qm_atmtab = new (atom *)[qm_natm];
	qm_bndtab = new (bond *)[qm_nbnd];

	mm_atmtab = new (atom *)[mm_natm];
	mm_bndtab = new (bond *)[mm_nbnd];
	
#ifdef ENABLE_SIMPLIFIED

	sf_atmtab = new (atom *)[sf_natm];
	sf_bndtab = new (bond *)[sf_nbnd];

#endif	// ENABLE_SIMPLIFIED
	
	// fill the tables.
	
	// BE SURE TO USE THE SAME RULES AS IN THE COUNTING STAGE!!!
	// BE SURE TO USE THE SAME RULES AS IN THE COUNTING STAGE!!!
	// BE SURE TO USE THE SAME RULES AS IN THE COUNTING STAGE!!!
	
	i32u qm_ac = 0;
	i32u mm_ac = 0;
	
#ifdef ENABLE_SIMPLIFIED
	
	i32u sf_ac = 0;
	
#endif	// ENABLE_SIMPLIFIED

	ita = mdl->GetAtomsBegin();
	while (ita != mdl->GetAtomsEnd())
	{
		bool ready = false;
		
#ifdef ENABLE_SIMPLIFIED
		
		if (!ready && (* ita).sf_atmtp != NOT_DEFINED)
		{
			sf_atmtab[sf_ac++] = & (* ita);
			ready = true;
		}
		
#endif	// ENABLE_SIMPLIFIED
		
		if (!ready && ((* ita).flags & ATOMFLAG_SELECTED))
		{
			qm_atmtab[qm_ac++] = & (* ita);
			ready = true;
		}
		else
		{
			mm_atmtab[mm_ac++] = & (* ita);
			ready = true;
		}
		
		if (!ready) { cout << "BUG: a2-counting error." << endl; exit(EXIT_FAILURE); }
		ita++;
	}
	
	i32u qm_bc = 0;
	i32u mm_bc = 0;
	
#ifdef ENABLE_SIMPLIFIED
	
	i32u sf_bc = 0;
	
#endif	// ENABLE_SIMPLIFIED

	itb = mdl->GetBondsBegin();
	while (itb != mdl->GetBondsEnd())
	{
		bool ready = false;
		
#ifdef ENABLE_SIMPLIFIED
		
		bool sff1 = ((* itb).atmr[0]->sf_atmtp != NOT_DEFINED);
		bool sff2 = ((* itb).atmr[1]->sf_atmtp != NOT_DEFINED);
		if (!ready && sff1 && sff2)
		{
			sf_bndtab[sf_bc++] = & (* itb);
			ready = true;
		}
		
#endif	// ENABLE_SIMPLIFIED
		
		bool qmf1 = ((* itb).atmr[0]->flags & ATOMFLAG_SELECTED);
		bool qmf2 = ((* itb).atmr[1]->flags & ATOMFLAG_SELECTED);
		if (!ready && qmf1 && qmf2)
		{
			qm_bndtab[qm_bc++] = & (* itb);
			ready = true;
		}
		else
		{
			mm_bndtab[mm_bc++] = & (* itb);
			ready = true;
		}
		
		if (!ready) { cout << "BUG: b2-counting error." << endl; exit(EXIT_FAILURE); }
		itb++;
	}
	
// these are quite bad errors; SHOULD NEVER HAPPEN!!!
// these are quite bad errors; SHOULD NEVER HAPPEN!!!
// these are quite bad errors; SHOULD NEVER HAPPEN!!!
	
	if ((i32s) qm_ac != qm_natm) cout << "WARNING : qm_natm mismatch!" << endl;
	if ((i32s) qm_bc != qm_nbnd) cout << "WARNING : qm_nbnd mismatch!" << endl;
	if ((i32s) mm_ac != mm_natm) cout << "WARNING : mm_natm mismatch!" << endl;
	if ((i32s) mm_bc != mm_nbnd) cout << "WARNING : mm_nbnd mismatch!" << endl;
	
#ifdef ENABLE_SIMPLIFIED
	
	if ((i32s) sf_ac != sf_natm) cout << "WARNING : sf_natm mismatch!" << endl;
	if ((i32s) sf_bc != sf_nbnd) cout << "WARNING : sf_nbnd mismatch!" << endl;
	
#endif	// ENABLE_SIMPLIFIED

}

void setup::CreateCurrentEngine(void)
{
	DiscardCurrentEngine();
	current_eng = CreateEngineByIndex(current_eng_index);
}

void setup::DiscardCurrentEngine(void)
{
	if (current_eng != NULL)
	{
		delete current_eng;
		current_eng = NULL;
	}
}

engine * setup::CreateEngineByIDNumber(i32u id)
{
	i32u index = 0;
	while (index < GetEngineCount())
	{
		if (GetEngineIDNumber(index) == id) break;
		else index++;
	}
	
	if (index < GetEngineCount())
	{
		return CreateEngineByIndex(index);
	}
	else
	{
		cout << "setup::CreateEngineByIDNumber() failed!" << endl;
		return NULL;
	}
}

/*################################################################################################*/

engine::engine(setup * p1, i32u p2)
{
	stp = p1;
	
	natm = stp->GetModel()->GetAtomCount();
	
	crd = new f64[natm * 3];
	
	if (p2 > 0) d1 = new f64[natm * 3];
	else d1 = NULL;
	
	if (p2 > 1) d2 = new f64[natm * natm * 9];
	else d2 = NULL;
}

engine::~engine(void)
{
	delete[] crd;
	
	if (d1 != NULL) delete[] d1;
	if (d2 != NULL) delete[] d2;
}

void engine::Check(i32u)
{
	const f64 delta = 0.000001;	// the finite difference step...
	
	Compute(1);
	f64 tmp1 = energy;
	
	f64 tmp2; f64 old;
	for (i32u n1 = 0;n1 < natm;n1++)
	{
		for (i32u n2 = 0;n2 < 3;n2++)
		{
			old = crd[n1 * 3 + n2];
			crd[n1 * 3 + n2] = old + delta;
			Compute(0); tmp2 = (energy - tmp1) / delta;
			crd[n1 * 3 + n2] = old;
			
			cout << n1 << ((char) ('x' + n2)) << " ";
			cout << "a = " << d1[n1 * 3 + n2] << " ";
			cout << "n = " << tmp2 << endl;
			
			if ((n1 % 5) == 4) cin >> old;
		}
	}
}

f64 engine::GetGradientVectorLength(void)
{
	f64 sum = 0.0;
	
	for (i32u n1 = 0;n1 < natm;n1++)
	{
		for (i32u n2 = 0;n2 < 3;n2++)
		{
			f64 tmp1 = d1[n1 * 3 + n2];
			sum += tmp1 * tmp1;
		}
	}
	
	return sqrt(sum);
}

/*################################################################################################*/

fGL value_VDWSurf(engine * eng, fGL * crd, fGL * grad)
{
	return eng->GetVDWSurf(crd, grad);
}

fGL value_ESP(engine * eng, fGL * crd, fGL * grad)
{
	return eng->GetESP(crd, grad);
}

fGL value_ElDens(engine * eng, fGL * crd, fGL * grad)
{
	return eng->GetElDens(crd, grad);
}

fGL value_Orbital(engine * eng, fGL * crd, fGL * grad)
{
	return eng->GetOrbital(crd, grad);
}

fGL value_OrbDens(engine * eng, fGL * crd, fGL * grad)
{
	return eng->GetOrbDens(crd, grad);
}

#ifdef ENABLE_SIMPLIFIED

// this #include macro is located here, because the eng1_sf class derived from the
// engine class and therefore it only can be introduced after the engine class itself...

#include "eng1_sf.h"

#endif	// ENABLE_SIMPLIFIED

void CopyCRD(model * p1, engine * p2, i32u p3)
{
	if (p3 >= p1->cs_vector.size())
	{
		cout << "BUG: cs overflow at CopyCRD() mdl->eng." << endl;
		exit(EXIT_FAILURE);
	}
	
	i32u n1 = 0; iter_al it1;
	for (it1 = p1->atom_list.begin();it1 != p1->atom_list.end();it1++)
	{
		const fGL * cdata = (* it1).GetCRD(p3);
		p2->crd[n1 * 3 + 0] = cdata[0];
		p2->crd[n1 * 3 + 1] = cdata[1];
		p2->crd[n1 * 3 + 2] = cdata[2];
		
		n1++;
	}
	
#ifdef ENABLE_SIMPLIFIED
	
	// so far, there is no need to copy any extra parameters in this direction...
	// so far, there is no need to copy any extra parameters in this direction...
	// so far, there is no need to copy any extra parameters in this direction...
	
#endif	// ENABLE_SIMPLIFIED

}

void CopyCRD(engine * p1, model * p2, i32u p3)
{
	if (p3 >= p2->cs_vector.size())
	{
		cout << "BUG: cs overflow at CopyCRD() eng->mdl." << endl;
		exit(EXIT_FAILURE);
	}
	
	i32u n1 = 0; iter_al it1;
	for (it1 = p2->atom_list.begin();it1 != p2->atom_list.end();it1++)
	{
		fGL x = p1->crd[n1 * 3 + 0];
		fGL y = p1->crd[n1 * 3 + 1];
		fGL z = p1->crd[n1 * 3 + 2];
		(* it1).SetCRD(p3, x, y, z);
		
		n1++;
	}
	
#ifdef ENABLE_SIMPLIFIED
	
	// in the simplified force field, we wish to display the main-chain peptide-unit dipoles.
	// check the engine object type, and copy the parameters if it matches. for graphics only!!!
	
	eng1_sf * p9 = dynamic_cast<eng1_sf *>(p1);
	if (p9 != NULL)
	{
		atom ** atmtab = p9->GetSetup()->GetSFAtoms();
		for (i32u n9 = 0;n9 < p9->bt3_vector.size();n9++)
		{
			atom * ref1 = atmtab[p9->l2g_sf[p9->bt3_vector[n9].atmi[1]]];
			atom * ref2 = atmtab[p9->l2g_sf[p9->bt3_vector[n9].atmi[2]]];
			
			bond tmpb1 = bond(ref1, ref2, bondtype());
			iter_bl itb1 = find(p2->GetBondsBegin(), p2->GetBondsEnd(), tmpb1);
			
			if (itb1 != p2->GetBondsEnd())
			{
				(* itb1).sf_pbdd = p9->bt3_vector[n9].pbdd;
			}
			else
			{
				cout << "WARNING : could not copy pbdd in CopyCRD()." << endl;
			}
		}
	}
	
#endif	// ENABLE_SIMPLIFIED

}

/*################################################################################################*/

// eof
