// OLDFILES.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "model.h"

#include <iomanip>
#include <strstream>
using namespace std;

/*################################################################################################*/

bool model::ReadStream_OLD(istream & p1, bool, bool)
{
	static const char message[] = "Hmmm. It seems that we are trying to read a VERY OLD version of this file.\nIt's not possible to verify the type and version of the file, so this\noperation may fail. Are you sure you wish to try (Y/N) ??? ";
	
	bool result = err->Question(message);
	if (!result) return false;
	
	vector<atom *> tmp1; i32s tmp2[2]; char tmp3; fGL tmp4[3];
	
	char buffer[1024];
	
#ifdef VERBOSE
cout << "model::ReadStream_OLD() : starting file input." << endl;
#endif	// VERBOSE

	while (p1.peek() != 'E')
	{
		p1 >> tmp2[0] >> tmp2[1];
		p1 >> tmp4[0] >> tmp4[1] >> tmp4[2] >> tmp2[0]; p1.get();
		
		element el = element(tmp2[1]);
		atom newatom(el, tmp4, cs_vector.size());
		AddAtom(newatom); tmp1.push_back(& atom_list.back());
		
		for (i32s n1 = 0;n1 < tmp2[0];n1++)
		{
			p1 >> tmp2[1] >> tmp3;
			if (tmp2[1] < (i32s) tmp1.size())
			{
				bond newbond(tmp1[tmp2[1]], tmp1[tmp1.size() - 1], bondtype(tmp3));
				AddBond(newbond);
			}
		}
		
		p1.getline(buffer, sizeof(buffer));
	}
	
	return true;
}

/*################################################################################################*/

bool model::ReadStream_v100(istream & p1, bool selected, bool skipheader)
{
	char buffer[1024];
	
	if (!skipheader)
	{
		// first we have to identify the file version, and then call the matching function,
		// if necessary. the very first file format did not use the "!Section" idea, so
		// we must skip the test if no such traces are seen...
		
		int version;
		if (p1.peek() != '!') version = -1;
		else
		{
			// now we expect to see "!Header" section.
			// next we will check the file type and version information...
			
			p1 >> buffer;		// read "!Header".
			p1 >> buffer;		// read the file type.
			p1 >> version;		// read the version number;
			
			p1.getline(buffer, sizeof(buffer));
		}
		
		// now look at the version number, and pick a correct function if needed.
		
		if (version < 100) return ReadStream_OLD(p1, selected, true);
		else if (version > 100)
		{
			static const char message[] = "Ooops. It seems that we are trying to read a FUTURE version of this file.\nIt's not sure if we can handle that. Are you sure you wish to try (Y/N) ??? ";
			
			bool result = err->Question(message);
			if (!result) return false;
		}
	}
	
	// the "!Header" section is now processed,
	// and the actual file input begins!!!
	
#ifdef VERBOSE
cout << "model::ReadStream() : starting file input." << endl;
#endif	// VERBOSE

	i32u crd_set_counter = 0;
	vector<atom *> ar_vector;
	
	// now we just read the sections, until we hit into "!End"-section and leave the loop...
	
	while (true)
	{
		p1 >> buffer;		// read the section string.
		
		if (!strcmp(buffer, "!End")) break;		// handle "!End" section.
		else if (!strcmp(buffer, "!Info"))		// handle "!Info" section.
		{
			int n1; p1 >> n1;
			
			int n2 = n1 - cs_vector.size();
			if (n2 > 0)
			{
				i32u n3 = cs_vector.size(); PushCRDSets((i32u) n2);
				for (i32u n4 = n3;n4 < cs_vector.size();n4++) SetCRDSetVisible(n4, true);
			}
			
#ifdef VERBOSE
cout << "handled \"!Info\" section, added " << n2 << " coordinate sets." << endl;
#endif	// VERBOSE

			p1.getline(buffer, sizeof(buffer));
			
			// TODO : read coordinate set descriptions (if there are some)...
			// TODO : read coordinate set descriptions (if there are some)...
			// TODO : read coordinate set descriptions (if there are some)...
		}
		else if (!strcmp(buffer, "!Atoms"))		// handle "!Atoms" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Atoms\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				i32s n2; p1 >> n2;	// atomic number
				
#ifdef VERBOSE
if (n1 != ar_vector.size()) cout << "WARNING : bad atom id numbers!!!" << endl;
#endif	// VERBOSE

				element tmp_element(n2);
				atom newatom(tmp_element, NULL, cs_vector.size());
				
				if (selected) newatom.flags |= ATOMFLAG_SELECTED;
				
				AddAtom(newatom);

				ar_vector.push_back(& atom_list.back());
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!Bonds"))		// handle "!Bonds" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Bonds\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u id1; p1 >> id1;	// 1st id number
				i32u id2; p1 >> id2;	// 2nd id number
				
				char bt; p1 >> bt;	// bond type
				
#ifdef VERBOSE
if (id1 >= ar_vector.size()) cout << "WARNING : bad atom id numbers!!!" << endl;
if (id2 >= ar_vector.size()) cout << "WARNING : bad atom id numbers!!!" << endl;
#endif	// VERBOSE

				bondtype tmp_bondtype(bt);
				bond newbond(ar_vector[id1], ar_vector[id2], tmp_bondtype);
				AddBond(newbond);
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else if (!strcmp(buffer, "!Coord"))		// handle "!Coord" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Coord\" section..." << endl;
if (crd_set_counter >= cs_vector.size()) cout << "ERROR : too many crd-sets!!!" << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				
				fGL xcrd; p1 >> xcrd;	// x-coordinate
				fGL ycrd; p1 >> ycrd;	// y-coordinate
				fGL zcrd; p1 >> zcrd;	// z-coordinate
				
#ifdef VERBOSE
if (n1 >= ar_vector.size()) cout << "ERROR : bad atom id numbers!!!" << endl;
#endif	// VERBOSE

				atom * ar = ar_vector[n1];
				ar->SetCRD(crd_set_counter, xcrd, ycrd, zcrd);
				
				p1.getline(buffer, sizeof(buffer));
			}
			
			crd_set_counter++;
		}
		else if (!strcmp(buffer, "!Charges"))		// handle "!Charges" section.
		{
			p1.getline(buffer, sizeof(buffer));
			
#ifdef VERBOSE
cout << "found \"!Charges\" section..." << endl;
#endif	// VERBOSE

			while (p1.peek() != '!')
			{
				i32u n1; p1 >> n1;	// id number
				f64 n2; p1 >> n2;	// charge
				
#ifdef VERBOSE
if (n1 >= ar_vector.size()) cout << "WARNING : bad atom id numbers!!!" << endl;
#endif	// VERBOSE

				atom * ar = ar_vector[n1];
				ar->charge = n2;
				
				p1.getline(buffer, sizeof(buffer));
			}
		}
		else						// handle an unknown section...
		{
			p1.getline(buffer, sizeof(buffer));
			
			while (p1.peek() != '!')
			{
				p1.getline(buffer, sizeof(buffer));
			}
		}
	}
	
	return true;
}

/*################################################################################################*/

// eof
