// UTILITY.H : many utility functions and classes...

// Copyright (C) 1998, 2002 Tommi Hassinen, Jarno Huuskonen, Geoff Hutchison.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "config.h"	// this is target-dependent...

#ifndef UTILITY_H
#define UTILITY_H

class superimpose;

/*################################################################################################*/

#include "atom.h"
#include "bond.h"

#include "model.h"

#include "conjgrad.h"

#include "v3d.h"

/*################################################################################################*/

/// This class can be used to superimpose coordinate sets.

class superimpose : public conjugate_gradient
{
	protected:
	
	i32s index[2];
	i32s counter;
	
	f64 value;
	f64 rot[3]; f64 drot[3];
	f64 loc[3]; f64 dloc[3];
	
	public:
	
	superimpose(i32s, i32s);
	virtual ~superimpose(void);
	
	f64 GetRMS(void) { return sqrt(value / (f64) counter); }
	void Compare(const f64 *, const f64 *, bool, f64 * = NULL);
	
	virtual void Transform(void) = 0;
};

/*################################################################################################*/


/**	An utility class (template) to release memory when program closes.
	
	If you have, say a table that you have to create when program starts, 
	and to destroy it when program closes. For example
	
	        int * table = new int[100];
	
	The memory will be automatically released if you attach an "cleaner" for it:
	
	        int * table = new int[100];
	        singleton_cleaner<int> table_cleaner(table);
*/

template<class TYPE1> class singleton_cleaner
{
	private:
	
	TYPE1 * instance;
	
	public:
	
	singleton_cleaner(TYPE1 * p1 = NULL) { instance = p1; }
	~singleton_cleaner(void) { if (instance) delete instance; }
	
	bool SetInstance(TYPE1 * p1) { if (instance) return false; instance = p1; return true; }
};

/*################################################################################################*/

/**	This function can be used to modify a 3D-vector object using a 4x4 transformation 
	matrix. Will first make a [1*4]x[4x4]-type matrix product and after that bring the 
	result back to 3D.
	
	For more complete documentation search for "##homogeneous coordinates"...
*/

void TransformVector(v3d<fGL> &, const fGL *);

/*################################################################################################*/

void DefineSecondaryStructure(model *);
f64 HBondEnergy(model *, i32s *, i32s *);

/*################################################################################################*/

#endif	// UTILITY_H

// eof
