// T2PROJECT.CPP

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "t2project.h"

#include "eng1_qm.h"
#include "eng1_mm.h"
#include "eng2_qm_mm.h"

#ifdef ENABLE_SIMPLIFIED
#include "eng1_sf.h"
#include "eng2_mm_sf.h"
#endif	// ENABLE_SIMPLIFIED

#include "intcrd.h"

#include "geomopt.h"
#include "moldyn.h"

#include "ribbon.h"

#include "plane.h"
#include "surface.h"

#include "color.h"
#include "views.h"

#include "filetrans.h"

#include <fstream>
#include <strstream>
using namespace std;

/*################################################################################################*/

glut_project::glut_project(glut_class_factory & p2) : project(p2)
{
}

glut_project::~glut_project(void)
{
}

void glut_project::KeyEvent(graphics_view * gv, char key, int, int)
{
	char buffer[1024];
	
	ifstream ifile;
	ofstream ofile;
	
	char input; i32s n1;
	
	static int setup_counter = 0;
	static readpdb_mdata * mdata = NULL;	// not quite safe...
	
	i32s type;
	static char objname[256];
	cs_param * param;
	
	fGL tmp1; fGL tmp2;
	i32s sz[3] = { 50, 50, 50 };
	fGL dm[3] = { 0.55, 0.50, 0.45 };
	
	cp_param cpp;
	cpp.np = 30; cpp.prj = this; cpp.ref = GetCurrentSetup()->GetCurrentEngine();
	
	cs_param csp1;
	csp1.np = sz; csp1.dim = dm;
	csp1.prj = this; csp1.ref = GetCurrentSetup()->GetCurrentEngine(); csp1.next = NULL;
	
	cs_param csp2a; cs_param csp2b;
	csp2a.np = sz; csp2a.dim = dm; csp2a.prj = this;
	csp2a.ref = GetCurrentSetup()->GetCurrentEngine(); csp2a.next = & csp2b;
	csp2b.np = sz; csp2b.dim = dm; csp2b.prj = this;
	csp2b.ref = GetCurrentSetup()->GetCurrentEngine(); csp2b.next = NULL;
	
	geomopt_param goparam;
	moldyn_param mdparam;
	
	engine * eng;
	
	switch (key)
	{
		case 'h':
		cout << "+ -       -> change the current element" << endl;
		cout << "< >       -> change the current bondtype" << endl;

		cout << "C         -> change color modes" << endl;
		cout << "r         -> change rendering modes (without updating view)" << endl;
		cout << "Q         -> change labels" << endl;

		cout << "O         -> open a file" << endl;
		cout << "S         -> save to a file" << endl;
		
		cout << "s         -> process a command string" << endl;
		
		cout << "H         -> hydrogens, add or remove" << endl;
		cout << "p         -> add (or \"push\") a new coordinate set" << endl;
//	cout << "0         -> superimpose the crd-sets and measure RMS-error" << endl;
		
		cout << "j         -> import PDB/ENT files" << endl;
		cout << "J         -> convert aa2sf" << endl;
		
		cout << "M         -> change the current_setup object" << endl;
		cout << "m         -> change the current_eng object" << endl;
		
		cout << "z         -> compare numerical and analytical gradient" << endl;
		cout << "E         -> compute energy [kJ/mol]" << endl;
		cout << "o         -> geometry optimization..." << endl;
		cout << "d         -> molecular dynamics..." << endl;
		cout << "R         -> random conformational search" << endl;

		cout << "t         -> build a protein ribbon model" << endl;
		
		cout << "T         -> add an energy-level diagram" << endl;
		
		cout << "8 9 0     -> make some nice planes/surfaces..." << endl;
		
		cout << "X         -> trajectory mode on/off" << endl;
		cout << "1 2 7     -> (in traj-mode) get prev/next/desired frame" << endl;
		cout << "5         -> (in traj-mode) play the trajectory file" << endl;
		break;
		
		case '+': ++element::current_element;
		cout << "current element = " << element::current_element.GetAtomicNumber();
		cout << " " << element::current_element.GetSymbol() << endl;
		break;
		
		case '-': --element::current_element;
		cout << "current element = " << element::current_element.GetAtomicNumber();
		cout << " " << element::current_element.GetSymbol() << endl;
		break;
		
		case '<': ++bondtype::current_bondtype;
		cout << "current bondtype = " << bondtype::current_bondtype << endl;
		break;
		
		case '>': --bondtype::current_bondtype;
		cout << "current bondtype = " << bondtype::current_bondtype << endl;
		break;
		
		case 'C':
		if (gv->colormode == & project::cm_element) gv->colormode = & project::cm_state;
		else if (gv->colormode == & project::cm_state) gv->colormode = & project::cm_element;
		UpdateAllGraphicsViews();
		break;
		
		case 'r':
		cout << "rendering = ";
		switch (gv->render)
		{
			case RENDER_NOTHING:
			gv->render = RENDER_WIREFRAME;
			cout << "RENDER_WIREFRAME";
			break;
			
			case RENDER_WIREFRAME:
			gv->render = RENDER_BALL_AND_STICK;
			cout << "RENDER_BALL_AND_STICK";
			break;
			
			case RENDER_BALL_AND_STICK:
			gv->render = RENDER_VAN_DER_WAALS;
			cout << "RENDER_VAN_DER_WAALS";
			break;
			
			case RENDER_VAN_DER_WAALS:
			gv->render = RENDER_CYLINDERS;
			cout << "RENDER_CYLINDERS";
			break;
			
			case RENDER_CYLINDERS:
			gv->render = RENDER_NOTHING;
			cout << "RENDER_NOTHING";
			break;
		}
		cout << endl;
		break;
		
		case 'Q':
		cout << "label = ";
		switch (gv->label)
		{
			case LABEL_NOTHING:
			gv->label = LABEL_INDEX;
			cout << "LABEL_INDEX";
			break;
			
			case LABEL_INDEX:
			gv->label = LABEL_CHARGE;
			cout << "LABEL_CHARGE";
			break;
			
			case LABEL_CHARGE:
			gv->label = LABEL_ELEMENT;
			cout << "LABEL_ELEMENT";
			break;
			
			case LABEL_ELEMENT:
			gv->label = LABEL_ATOMTYPE;
			cout << "LABEL_ATOMTYPE";
			break;
			
			case LABEL_ATOMTYPE:
			gv->label = LABEL_BONDTYPE;
			cout << "LABEL_BONDTYPE";
			break;
			
			case LABEL_BONDTYPE:
			gv->label = LABEL_NOTHING;
			cout << "LABEL_NOTHING";
			break;
		}
		cout << endl;
		break;
		
		case 'O':
		cout << "name of file to open ??? "; cin.getline(buffer, sizeof(buffer));
		ifile.open(buffer, ios::in); ReadStream(ifile); ifile.close();
		UpdateAllGraphicsViews();
		break;
		
		case 'S':
		cout << "name of file to save ??? "; cin.getline(buffer, sizeof(buffer));
		ofile.open(buffer, ios::out); WriteStream(ofile); ofile.close();
		break;
		
		case 's':
		cout << "command string ??? "; cin.getline(buffer, sizeof(buffer));
		ProcessCommandString(gv, buffer);
		break;
		
		case 'H':
		cout << "hydrogens, add or remove ? (a/r) ";
		cin >> input;
		
		if (input == 'a')
		{
			cout << "adding hydrogens..." << endl;
			AddHydrogens(); UpdateAllGraphicsViews();
		}
		
		if (input == 'r')
		{
			cout << "removing hydrogens..." << endl;
			RemoveHydrogens(); UpdateAllGraphicsViews();
		}
		break;
		
		case 'p':
		PushCRDSets(1); n1 = cs_vector.size() - 1;
		while (n1 > 0) { CopyCRDSet(n1 - 1, n1); n1 -= 1; }
		cs_vector.back()->visible = true;
		UpdateAllGraphicsViews();
		break;
		
	/*	case '0': t2a = 0.0;
		for (t1a = 0;t1a < ((i32s) cs_vector.size()) - 1;t1a++)
		{
			// the 'p'-command above adds new sets to front, so we will
			// superimpose to the last one here (which was saved first).
			
			superimpose si(this, ((i32s) cs_vector.size()) - 1, t1a);
			for (t1b = 0;t1b < 100;t1b++) si.TakeCGStep(conjugate_gradient::Newton2An);
			
			t2b = si.GetRMS(); t2a += t2b; si.Transform();
			cout << t1a << " RMS-distance = " << t2b << endl;
		}
		cout << "average = " << (t2a / (((fGL) cs_vector.size()) - 1.0)) << endl;
		UpdateAllGraphicsViews();
		break;	*/
		
		case 'j':
		cout << "import ENT: filename ??? "; cin >> buffer;
		cout << "give the index of chain to be imported, or -1 to import all ??? "; cin >> n1;
		
		mdata = readpdb_ReadMData(buffer);
		readpdb_ReadData(buffer, mdata, n1);
		
		UpdateAllGraphicsViews();
		break;
		
		case 'J':
		if (mdata == NULL) { cout << "you should import a PDB file first!" << endl; return; }
		
		aa2sf_ConvertAll(mdata);
		UpdateAllGraphicsViews();
		break;
		
		case 'M':
		delete current_setup;
		cout << "current_setup = ";
		switch (setup_counter)
		{
			case 0:
			cout << "setup1_qm (quantum mechanics)";
			current_setup = new setup1_qm(this);
			setup_counter = 1;
			break;

			case 1:
			cout << "setup2_qm_mm (mixed qm/mm)";
			current_setup = new setup2_qm_mm(this);
			setup_counter = 2;
			break;
			
#ifdef ENABLE_SIMPLIFIED
			
			case 2:
			cout << "setup1_sf (simplified forcefield)";
			current_setup = new setup1_sf(this);
			setup_counter = 3;
			break;
			
			case 3:
			cout << "setup2_mm_sf (mixed mm/sf)";
			current_setup = new setup2_mm_sf(this);
			setup_counter = 4;
			break;
			
#endif	// ENABLE_SIMPLIFIED
			
			default:
			cout << "setup1_mm (molecular mechanics)";
			current_setup = new setup1_mm(this);
			setup_counter = 0;
		}
		cout << endl;
		cout << "current_eng = " << GetCurrentSetup()->GetEngineName(GetCurrentSetup()->current_eng_index) << endl;
		break;
		
		case 'm':
		GetCurrentSetup()->current_eng_index++;
		if (GetCurrentSetup()->current_eng_index >= GetCurrentSetup()->GetEngineCount()) GetCurrentSetup()->current_eng_index = 0;
		cout << "current_eng = " << GetCurrentSetup()->GetEngineName(GetCurrentSetup()->current_eng_index) << endl;
		break;
		
		case 'z':
	/*	{ //LOCAL
		file_trans ft;
		OEMol * oem = ft.CopyAll(this);
		
	//	oem->SetOutputType(GHEMICAL);
	//	cout << (* oem) << endl;
	
		oem->AddHydrogens(false, false);
		
		ft.Synchronize();
		UpdateAllGraphicsViews();
		} //LOCAL	*/
		
		eng = GetCurrentSetup()->CreateEngineByIndex(GetCurrentSetup()->current_eng_index);
		CopyCRD(this, eng, 0); eng->Check(1); delete eng;
		break;
		
		case 'E':
		DoEnergy();
		break;
		
		case 'o':
		goparam.treshold_nsteps = 500;
		DoGeomOpt(goparam);
		break;
		
		case 'd':
		mdparam.nsteps = 10000;
		mdparam.temperature = 300.0;
		mdparam.timestep = 0.5;
		DoMolDyn(mdparam);
		break;
		
		case 'R':
		i32s cycles; i32s optsteps;
		cout << "give number of cycles (100 would be ok) or -1 to cancel : "; cin >> cycles; if (cycles < 0) return;
		cout << "give number of optimization steps in a cycle (250 would be ok) : "; cin >> optsteps;
		DoRandomSearch(cycles, optsteps);
		
	/*	i32s divisions; i32s optsteps;
		cout << "give number of divisions (for example 6 divisions ->\n360 deg / 6 = 60 deg) or -1 to cancel : "; cin >> divisions; if (divisions < 0) return;
		cout << "give number of optimization steps in a cycle (250 would be ok,\nor 0 to skip optimization) : "; cin >> optsteps;
		DoSystematicSearch(divisions, optsteps);	*/
		
	/*	i32s n_init_steps; i32s n_simul_steps; i32s optsteps;
		cout << "give number of init and simul steps (10 100 would be ok)\nor -1 to cancel : "; cin >> n_init_steps; cin >> n_simul_steps; if (n_init_steps < 0 || n_simul_steps < 0) return;
		cout << "give number of optimization steps in a cycle (250 would be ok,\nor 0 to skip optimization) : "; cin >> optsteps;
		DoMonteCarloSearch(n_init_steps, n_simul_steps, optsteps);	*/
		break;
		
		case 't':
		UpdateChains();
		DefineSecondaryStructure(this);
		
		{	// USE LOCAL VARIABLES HERE...
			vector<chn_info> & ci_vector = (* this->ref_civ);
			for (i32u n1 = 0;n1 < ci_vector.size();n1++)
			{
				if (ci_vector[n1].GetType() != chn_info::amino_acid) continue;
				if (ci_vector[n1].GetLength() < 3) continue;
				
				AddObject(new ribbon(this, gv->colormode, n1, 4));	// min. order is 2!!!
			}
		}
		
	//	UpdateAllGraphicsViews();
		break;

		case 'T':
		AddEnLevDiagView();
		break;
		
		case '8':
if (!GetCurrentSetup()->GetCurrentEngine()) { cout << "calculate energy first!" << endl; break; }

		cout << "plane objects:" << endl;
		cout << "--------------" << endl;
		cout << "1 - esp" << endl;
		cout << "2 - eldens" << endl;
		cout << "3 - mo" << endl;
		cout << "4 - mo-dens" << endl;
		cout << "give the type of plane object (or -1 to cancel) ??? "; cin >> type;
		if (type < 1) break;
		
		cout << "give dimension of the plane [nm] (1.0 might be ok) ??? "; cin >> tmp1;
		
		switch (type)
		{
			case 1:
			cpp.dim = tmp1; cpp.transparent = true;
			cpp.vf = (ValueFunction *) value_ESP; cpp.cf = (ColorFunction *) GetRBRange1;
			cpp.value = 138.9354518 * 1.0; cpp.alpha = 0.75;
			strcpy(objname, "ESP-");
			break;
			
			case 2:
			cpp.dim = tmp1; cpp.transparent = true;
			cpp.vf = (ValueFunction *) value_ElDens; cpp.cf = (ColorFunction *) GetGreenColor;
			cpp.value = 0.05; cpp.alpha = 0.75;
			strcpy(objname, "E-density ");
			break;
			
			case 3:
			cpp.dim = tmp1; cpp.transparent = true;
			cpp.vf = (ValueFunction *) value_Orbital; cpp.cf = (ColorFunction *) GetRBRange1;
			cpp.value = 0.05; cpp.alpha = 0.75;
			strcpy(objname, "molecular orbital ");
			break;
			
			case 4:
			cpp.dim = tmp1; cpp.transparent = true;
			cpp.vf = (ValueFunction *) value_OrbDens; cpp.cf = (ColorFunction *) GetGreenColor;
			cpp.value = 0.0125; cpp.alpha = 0.75;
			strcpy(objname, "MO-density ");
			break;
			
			default:
			cout << "unknown type!!!" << endl;
			return;
		}
		AddObject(new color_plane_object(ol_static(), cpp, objname));
	//	UpdateAllGraphicsViews();
		break;
		
		case '9':
if (!GetCurrentSetup()->GetCurrentEngine()) { cout << "calculate energy first!" << endl; break; }

		cout << "surface objects:" << endl;
		cout << "--------------" << endl;
		cout << "1 - esp" << endl;
		cout << "2 - esp-colored vdw" << endl;
		cout << "3 - e-dens" << endl;
		cout << "4 - mo" << endl;
		cout << "5 - mo-dens" << endl;
		cout << "give the type of surface object (or -1 to cancel) ??? "; cin >> type;
		if (type < 1) break;
		
		cout << "give dimension of the surface [nm] (2.5 might be ok) ??? ";
		cin >> tmp2; dm[0] = dm[1] = dm[2] = tmp2;

		cout << "choose the type of surface : solid or wireframe ? (s/w) ";
		cin >> input; csp1.wireframe = csp2a.wireframe = csp2b.wireframe = (input == 'w' || input == 'W');
		
		switch (type)
		{
			case 1:
			tmp1 = 138.9354518 / 4.0;
		csp2a.transparent = !true;
		csp2a.vf1 = (ValueFunction *) value_ESP; csp2a.vf2 = (ValueFunction *) GetUnity;
		csp2a.cvalue = 1.0; csp2a.alpha = 0.25; csp2a.toler = 1.0e-6 * tmp1; csp2a.maxc = 50;
			csp2a.cf = (ColorFunction *) GetRedColor; csp2a.svalue = +tmp1;
		csp2b.transparent = !true;
		csp2b.vf1 = (ValueFunction *) value_ESP; csp2b.vf2 = (ValueFunction *) GetUnity;
		csp2b.cvalue = 1.0; csp2b.alpha = 0.25; csp2b.toler = 1.0e-6 * tmp1; csp2b.maxc = 50;
			csp2b.cf = (ColorFunction *) GetBlueColor; csp2b.svalue = -tmp1;
			param = & csp2a; strcpy(objname, "ESP-");
			break;
			
			case 2:
			tmp1 = 138.9354518 / 2.0;
		csp1.transparent = true; csp1.cf = (ColorFunction *) GetRBRange1;
		csp1.vf1 = (ValueFunction *) value_VDWSurf; csp1.vf2 = (ValueFunction *) value_ESP;
		csp1.svalue = 1.0; csp1.cvalue = tmp1; csp1.alpha = 0.50; csp1.toler = 1.0e-6 * tmp1; csp1.maxc = 50;
			param = & csp1; strcpy(objname, "ESP-colored VDW-");
			break;
			
			case 3:
			tmp1 = 0.05;
		csp1.transparent = true; csp1.cf = (ColorFunction *) GetGreenColor;
		csp1.vf1 = (ValueFunction *) value_ElDens; csp1.vf2 = (ValueFunction *) GetUnity;
		csp1.svalue = tmp1; csp1.cvalue = 1.0; csp1.alpha = 0.50; csp1.toler = 1.0e-6 * tmp1; csp1.maxc = 250;
			param = & csp1; strcpy(objname, "E-density ");
			break;
			
			case 4:
			tmp1 = 0.05;
		csp2a.transparent = !true;
		csp2a.vf1 = (ValueFunction *) value_Orbital; csp2a.vf2 = (ValueFunction *) GetUnity;
		csp2a.cvalue = 1.0; csp2a.alpha = 0.50; csp2a.toler = 1.0e-6 * tmp1; csp2a.maxc = 250;
			csp2a.cf = (ColorFunction *) GetRedColor; csp2a.svalue = +tmp1;
		csp2b.transparent = !true;
		csp2b.vf1 = (ValueFunction *) value_Orbital; csp2b.vf2 = (ValueFunction *) GetUnity;
		csp2b.cvalue = 1.0; csp2b.alpha = 0.50; csp2b.toler = 1.0e-6 * tmp1; csp2b.maxc = 250;
			csp2b.cf = (ColorFunction *) GetBlueColor; csp2b.svalue = -tmp1;
			param = & csp2a; strcpy(objname, "MO ");
			break;
			
			case 5:
			tmp1 = 0.0125;
		csp1.transparent = true; csp1.cf = (ColorFunction *) GetGreenColor;
		csp1.vf1 = (ValueFunction *) value_OrbDens; csp1.vf2 = (ValueFunction *) GetUnity;
		csp1.svalue = tmp1; csp1.cvalue = 1.0; csp1.alpha = 0.50; csp1.toler = 1.0e-6 * tmp1; csp1.maxc = 250;
			param = & csp1; strcpy(objname, "MO-density ");
			break;
			
			default:
			cout << "unknown type!!!" << endl;
			return;
		}
		AddObject(new color_surface_object(ol_static(), * param, objname));
	//	UpdateAllGraphicsViews();
		break;
		
		case '0':
if (!GetCurrentSetup()->GetCurrentEngine()) { cout << "calculate energy first!" << endl; break; }

		cout << "volume objects:" << endl;
		cout << "--------------" << endl;
		cout << "1 - esp" << endl;
		cout << "2 - esp-colored vdw" << endl;
		cout << "give the type of volume object (or -1 to cancel) ??? "; cin >> type;
		if (type < 1) break;
		
		cout << "give dimension of the volume [nm] (1.0 might be ok) ??? "; cin >> tmp1;
		
		switch (type)
		{
			case 1:
		cpp.dim = tmp1; cpp.transparent = true;
		cpp.vf = (ValueFunction *) value_ESP; cpp.cf = (ColorFunction *) GetRBRange2;
		cpp.value = 138.9354518 * 1.0; cpp.alpha = 0.25;
			strcpy(objname, "ESP-");
			break;
			
			case 2:
		cpp.dim = tmp1; cpp.transparent = true;
		cpp.vf = (ValueFunction *) value_Orbital; cpp.cf = (ColorFunction *) GetRBRange2;
		cpp.value = 0.05; cpp.alpha = 0.25;
			strcpy(objname, "MO ");
			break;
			
			default:
			cout << "unknown type!!!" << endl;
			return;
		}
		AddObject(new volume_rendering_object(ol_static(), cpp, 20, tmp1 / 2.0, (* gv->cam), objname));
	//	UpdateAllGraphicsViews();
		break;
		
		case 'X':
		if (GetTrajectoryFile() != NULL)
		{
			cout << "traj mode off" << endl;
			CloseTrajectory();
		}
		else
		{
			cout << "traj mode on" << endl;
			cout << "traj-file name ??? "; cin >> buffer;
			OpenTrajectory(buffer);
		}
		break;
		
		case '1':
		case '2':
		case '7':
		if (GetTrajectoryFile() != NULL)
		{
			i32s frame = GetCurrentFrame();
			
			switch (key)
			{
				case '1':
				frame--; if (frame < 0) frame = 0;
				break;
				
				case '2':
				frame++; if (frame == GetTotalFrames()) frame = GetTotalFrames() - 1;
				break;
				
				default:
				cout << "give the frame (in correct range) ??? "; cin >> frame;
			}
			
			cout << "frame = " << frame << endl;
			
			SetCurrentFrame(frame);
			ReadFrame();
			
			UpdateAllGraphicsViews();
		} else cout << "in traj-mode only!!!" << endl;
		break;
		
		case '5':
		if (GetTrajectoryFile() != NULL)
		{
			for (i32s loop = 0;loop < GetTotalFrames();loop++)
			{
				SetCurrentFrame(loop);
				ReadFrame();
				
				UpdateAllGraphicsViews(true);
			}
		} else cout << "in traj-mode only!!!" << endl;
		break;
		
		default:
		cout << "unknown command '" << key << "'" << endl;
	}
}

void glut_project::event_SelectedObjectChanged(void)
{
	UpdateAllGraphicsViews();	// this will update the object information...
}

void glut_project::GeomOptGetParam(geomopt_param & param)
{
	cout << "how many cg-steps (" << param.treshold_nsteps << " might be Ok, -1 = cancel) ??? ";
	cin >> param.treshold_nsteps;
	
	param.enable_nsteps = true;
	param.enable_grad = false; param.treshold_grad = 0.0;		// skip the other tests...
	param.enable_delta_e = false; param.treshold_delta_e = 0.0;	// skip the other tests...
	
	param.confirm = !(param.treshold_nsteps < 0);
}

#define START 10000	// see project.cpp!!!!!!!!

void glut_project::MolDynGetParam(moldyn_param & param)
{
	cout << "how many md-steps for simulation (" << param.nsteps << " might be Ok, -1 = cancel) ??? ";
	cin >> param.nsteps;
	
	param.confirm = !(param.nsteps < 0);
	if (!param.confirm) return;
	
	cout << "there are also " << START << " heating steps, total steps = " << (START + param.nsteps) << endl;
	cout << "the trajectory file will contain " << (param.nsteps / 100) << " frames!!!" << endl;
	
	cout << "simulation temperature [K] (" << param.temperature << " might be Ok) ??? ";
	cin >> param.temperature;
	
	cout << "time-step length [fs] (" << param.timestep << " might be Ok) ??? ";
	cin >> param.timestep;
	
	cout << "traj-file name ??? ";
	cin >> param.filename;
}

/*################################################################################################*/

// eof
