// SETUP_DRUID.CPP

// Copyright (C) 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#include "setup_druid.h"

#include "eng1_qm.h"
#include "eng1_qm.h"
#include "eng2_qm_mm.h"

#ifdef ENABLE_SIMPLIFIED
#include "eng1_sf.h"
#include "eng2_mm_sf.h"
#endif	// ENABLE_SIMPLIFIED

#include <gnome.h>

#include <iostream>
#include <strstream>
using namespace std;

/*################################################################################################*/

setup_druid::setup_druid(gnome_project * p1) : glade_dialog("glade/setup_druid.glade")
{
	prj = p1;
	
	window = glade_xml_get_widget(xml, "setup_druid_window");
	if (window == NULL) { cout << "setup_druid : glade_xml_get_widget() failed!!!" << endl; return; }
	
	pageinfo = NOT_DEFINED;
	druid = glade_xml_get_widget(xml, "setup_druid");
	
	page1 = glade_xml_get_widget(xml, "page1");
	page2[0] = glade_xml_get_widget(xml, "page2a");
	page2[1] = glade_xml_get_widget(xml, "page2b");
	page2[2] = glade_xml_get_widget(xml, "page2c");
	page2[3] = glade_xml_get_widget(xml, "page2d");
	page2[4] = glade_xml_get_widget(xml, "page2e");
	
	page1_rbuttons[0] = glade_xml_get_widget(xml, "allqm_rbutton");
	page1_rbuttons[1] = glade_xml_get_widget(xml, "allmm_rbutton");
	page1_rbuttons[2] = glade_xml_get_widget(xml, "qmmm_rbutton");
	page1_rbuttons[3] = glade_xml_get_widget(xml, "allsf_rbutton");
	page1_rbuttons[4] = glade_xml_get_widget(xml, "mmsf_rbutton");
	
	// fill the optionmenu widgets and some other settings at druid pages...
	// fill the optionmenu widgets and some other settings at druid pages...
	// fill the optionmenu widgets and some other settings at druid pages...
	
	GtkWidget * optmenu; GtkWidget * submenu; GtkWidget * item; i32u eng_index;
	GtkWidget * saved_optmenus[5];	// these are for setting the default engine...
	
	// setup1_qm
	
	optmenu = glade_xml_get_widget(xml, "optionmenu_allqm_eng");
	submenu = gtk_menu_new(); saved_optmenus[0] = optmenu;
	
	for (eng_index = 0; eng_index < setup1_qm::static_GetEngineCount(); eng_index++)
	{
		item = gtk_menu_item_new_with_label(setup1_qm::static_GetEngineName(eng_index));
		gtk_widget_show(item); gtk_menu_append(GTK_MENU(submenu), item);
	}
	
	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), submenu);
	gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), prj->GetCurrentSetup()->current_eng_index);
	
	// setup1_mm

	optmenu = glade_xml_get_widget(xml, "optionmenu_allmm_eng");
	submenu = gtk_menu_new(); saved_optmenus[1] = optmenu;
	
	for (eng_index = 0; eng_index < setup1_mm::static_GetEngineCount(); eng_index++)
	{
		item = gtk_menu_item_new_with_label(setup1_mm::static_GetEngineName(eng_index));
		gtk_widget_show(item); gtk_menu_append(GTK_MENU(submenu), item);
	}
	
	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), submenu);
	
	// setup2_qm_mm
	
	optmenu = glade_xml_get_widget(xml, "optionmenu_qmmm_eng");
	submenu = gtk_menu_new(); saved_optmenus[2] = optmenu;
	
	for (eng_index = 0; eng_index < setup2_qm_mm::static_GetEngineCount(); eng_index++)
	{
		item = gtk_menu_item_new_with_label(setup2_qm_mm::static_GetEngineName(eng_index));
		gtk_widget_show(item); gtk_menu_append(GTK_MENU(submenu), item);
	}
	
	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), submenu);
	
	// setup1_sf
	// setup2_mm_sf
	// NOTHING IS DONE FOR THESE SINCE THERE ARE ONLY A SINGLE ENG AVAILABLE!!!
	
	// detect the type of model::current_setup object, and setup the default values...
	// detect the type of model::current_setup object, and setup the default values...
	// detect the type of model::current_setup object, and setup the default values...
	
	setup1_qm * su_allqm = dynamic_cast<setup1_qm *>(prj->GetCurrentSetup());
	setup1_mm * su_allmm = dynamic_cast<setup1_mm *>(prj->GetCurrentSetup());
	setup2_qm_mm * su_qmmm = dynamic_cast<setup2_qm_mm *>(prj->GetCurrentSetup());
	
#ifdef ENABLE_SIMPLIFIED
	
	setup1_sf * su_allsf = dynamic_cast<setup1_sf *>(prj->GetCurrentSetup());
	setup2_mm_sf * su_mmsf = dynamic_cast<setup2_mm_sf *>(prj->GetCurrentSetup());
	
#endif	// ENABLE_SIMPLIFIED

	// the defaults for the subpages...
	// the defaults for the subpages...
	// the defaults for the subpages...
	
	if (su_allqm == NULL)
	{
		GtkWidget * entry_charge = glade_xml_get_widget(xml, "entry_allqm_charge");	// total charge
		gtk_entry_set_text(GTK_ENTRY(entry_charge), "0.0");
		
		GtkWidget * entry_spin = glade_xml_get_widget(xml, "entry_allqm_spin");		// spin multiplicity
		gtk_entry_set_text(GTK_ENTRY(entry_spin), "1");
	}
	
	if (su_allmm == NULL)
	{
		GtkWidget * entry_dimx = glade_xml_get_widget(xml, "entry_allmm_dimx");		// pbc dim x
		gtk_entry_set_text(GTK_ENTRY(entry_dimx), "1.500");
		
		GtkWidget * entry_dimy = glade_xml_get_widget(xml, "entry_allmm_dimy");		// pbc dim y
		gtk_entry_set_text(GTK_ENTRY(entry_dimy), "1.500");
		
		GtkWidget * entry_dimz = glade_xml_get_widget(xml, "entry_allmm_dimz");		// pbc dim z
		gtk_entry_set_text(GTK_ENTRY(entry_dimz), "1.500");
	}
	
	if (su_qmmm == NULL)
	{
	}
	
#ifdef ENABLE_SIMPLIFIED

	if (su_allsf == NULL)
	{
	}
	
	if (su_mmsf == NULL)
	{
	}
	
#endif	// ENABLE_SIMPLIFIED

	// the settings specific to the model::current_setup object...
	// the settings specific to the model::current_setup object...
	// the settings specific to the model::current_setup object...

	if (su_allqm != NULL)
	{
		gtk_option_menu_set_history(GTK_OPTION_MENU(saved_optmenus[0]), su_allqm->current_eng_index);
		
		page1_current_rb = page1_rbuttons[0];
		
	//	char buffer[64];
		
		GtkWidget * entry_charge = glade_xml_get_widget(xml, "entry_allqm_charge");	// total charge
		gtk_entry_set_text(GTK_ENTRY(entry_charge), "???");
		
		GtkWidget * entry_spin = glade_xml_get_widget(xml, "entry_allqm_spin");		// spin multiplicity
		gtk_entry_set_text(GTK_ENTRY(entry_spin), "???");

/*		
GtkWidget * entry_tc = glade_xml_get_widget(xml, "entry_tc");		// total charge
ostrstream str_tc(buffer, sizeof(buffer));
str_tc.setf(ios::showpos); str_tc << owner->GetTotalCharge() << ends;
gtk_entry_set_text(GTK_ENTRY(entry_tc), buffer);	*/
	}
	else if (su_allmm != NULL)
	{
		gtk_option_menu_set_history(GTK_OPTION_MENU(saved_optmenus[1]), su_allmm->current_eng_index);
		
		page1_current_rb = page1_rbuttons[1];
		
		GtkWidget * entry_dimx = glade_xml_get_widget(xml, "entry_allmm_dimx");		// pbc dim x
		gtk_entry_set_text(GTK_ENTRY(entry_dimx), "???");
		
		GtkWidget * entry_dimy = glade_xml_get_widget(xml, "entry_allmm_dimy");		// pbc dim y
		gtk_entry_set_text(GTK_ENTRY(entry_dimy), "???");
		
		GtkWidget * entry_dimz = glade_xml_get_widget(xml, "entry_allmm_dimz");		// pbc dim z
		gtk_entry_set_text(GTK_ENTRY(entry_dimz), "???");
	}
	else if (su_qmmm != NULL)
	{
		gtk_option_menu_set_history(GTK_OPTION_MENU(saved_optmenus[2]), su_qmmm->current_eng_index);
		
		page1_current_rb = page1_rbuttons[2];
	}
	
#ifdef ENABLE_SIMPLIFIED

	else if (su_allsf != NULL)
	{
		// saved_optmenus[3] is not initialized yet...
		page1_current_rb = page1_rbuttons[3];
	}
	else if (su_mmsf != NULL)
	{
		// saved_optmenus[4] is not initialized yet...
		page1_current_rb = page1_rbuttons[4];
	}
	
#endif	// ENABLE_SIMPLIFIED

	else
	{
		prj->err->ErrorMessage("BUG: Bad cast was detected at setup_druid ctor.\nThe program will terminate.");
		exit(EXIT_FAILURE);
	}
	
	// ok, continue the druid setup...
	// ok, continue the druid setup...
	// ok, continue the druid setup...
	
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(page1_current_rb), TRUE);
	
	gnome_druid_set_page(GNOME_DRUID(druid), GNOME_DRUID_PAGE(page1));
	gnome_druid_set_buttons_sensitive(GNOME_DRUID(druid), FALSE, TRUE, TRUE);
	gnome_druid_set_show_finish(GNOME_DRUID(druid), FALSE);
	
	// connect the handlers...
	
	glade_xml_signal_connect_data(xml, "on_window_destroy", (GtkSignalFunc) handler_Destroy, (gpointer) this);
	
	glade_xml_signal_connect_data(xml, "on_p1_rbutton_clicked", (GtkSignalFunc) handler_Page1RadioButton, (gpointer) this);
	
	glade_xml_signal_connect_data(xml, "on_dp_back", (GtkSignalFunc) handler_DruidButtonBack, (gpointer) this);
	glade_xml_signal_connect_data(xml, "on_dp_next", (GtkSignalFunc) handler_DruidButtonNext, (gpointer) this);
	glade_xml_signal_connect_data(xml, "on_dp_cancel", (GtkSignalFunc) handler_DruidButtonCancel, (gpointer) this);
	glade_xml_signal_connect_data(xml, "on_dp_finish", (GtkSignalFunc) handler_DruidButtonFinish, (gpointer) this);
	
	gtk_widget_show(window);	// MODELESS
}

setup_druid::~setup_druid(void)
{
}

void setup_druid::handler_Destroy(GtkWidget *, gpointer data)
{
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_Destroy() : ref = " << ref << endl;
	
	delete ref;	// the object was created with new!!!
}

void setup_druid::handler_Page1RadioButton(GtkWidget * rb, gpointer data)
{
// the toggle-button-style system is not actually needed, since the rbutton groups seem to work just fine.
// however, the "old"-style code is still here, commented out (similar is still used in mousetool buttons).
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
//	static bool block_handler = false;	// prevent cascading button events...
//	if (block_handler) return;		// also look at gnome_mdi_app::ToolButtonHandler().
	
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_Page1RadioButton() : ref = " << ref << endl;
	
//	block_handler = true;
//	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ref->page1_current_rb), FALSE);
	
	for (int n1 = 0;n1 < 5;n1++)
	{
		if (ref->page1_rbuttons[n1] == rb)
		{
	//		cout << "the widget " << n1 << " matches!" << endl;
			
			ref->page1_current_rb = ref->page1_rbuttons[n1];
			break;
		}
	}
	
//	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ref->page1_current_rb), TRUE);
//	block_handler = false;
}

gboolean setup_druid::handler_DruidButtonBack(GnomeDruidPage *, gpointer, gpointer data)
{
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_DruidButtonBack() : ref = " << ref << endl;
	
	// only page2 can send a "back" event...
	// ...so, just get back to page1.
	
	gnome_druid_set_page(GNOME_DRUID(ref->druid), GNOME_DRUID_PAGE(ref->page1));
	gnome_druid_set_buttons_sensitive(GNOME_DRUID(ref->druid), FALSE, TRUE, TRUE);
	gnome_druid_set_show_finish(GNOME_DRUID(ref->druid), FALSE);
	
	return TRUE;
}

gboolean setup_druid::handler_DruidButtonNext(GnomeDruidPage *, gpointer, gpointer data)
{
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_DruidButtonNext() : ref = " << ref << endl;
	
	// only page1 can send a "next" event...
	// ...so, we will detect which rbutton was active at page1 and set the corresponding page2.
	
	int n1;
	for (n1 = 0;n1 < 5;n1++)
	{
		if (ref->page1_rbuttons[n1] == ref->page1_current_rb) break;
	}
	
	if (n1 < 0 || n1 >= 5) cout << "oops! could not select page2." << endl;
	else
	{
		ref->pageinfo = n1;
		
		gnome_druid_set_page(GNOME_DRUID(ref->druid), GNOME_DRUID_PAGE(ref->page2[n1]));
		gnome_druid_set_buttons_sensitive(GNOME_DRUID(ref->druid), TRUE, TRUE, TRUE);
		gnome_druid_set_show_finish(GNOME_DRUID(ref->druid), TRUE);
	}
	
	return TRUE;
}

gboolean setup_druid::handler_DruidButtonFinish(GnomeDruidPage *, gpointer, gpointer data)
{
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_DruidButtonFinish() : ref = " << ref << endl;
	
	// read in and process the user's settings...
	// read in and process the user's settings...
	// read in and process the user's settings...
	
	GtkWidget * optmenu; GtkWidget * submenu; GtkWidget * item; int index;
	
	switch (ref->pageinfo)
	{
		case 0:
		optmenu = glade_xml_get_widget(ref->xml, "optionmenu_allqm_eng");
		submenu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));
		item = gtk_menu_get_active(GTK_MENU(submenu));
		
		index = g_list_index(GTK_MENU_SHELL(submenu)->children, item);
		
/*		owner->DiscardCurrentEng();
		
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		
		if (owner->CheckEngSettings() == false) return;	*/
		
		delete ref->prj->current_setup;						// get rid of the old setup...
		ref->prj->current_setup = new setup1_qm(ref->prj);			// ...and make a new one.
		
		ref->prj->GetCurrentSetup()->current_eng_index = index;			// set the requested engine.
		break;
		
		case 1:
		optmenu = glade_xml_get_widget(ref->xml, "optionmenu_allmm_eng");
		submenu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));
		item = gtk_menu_get_active(GTK_MENU(submenu));
		
		index = g_list_index(GTK_MENU_SHELL(submenu)->children, item);
		
		delete ref->prj->current_setup;						// get rid of the old setup...
		ref->prj->current_setup = new setup1_mm(ref->prj);			// ...and make a new one.
		
		ref->prj->GetCurrentSetup()->current_eng_index = index;			// set the requested engine.
		break;
		
		case 2:
		optmenu = glade_xml_get_widget(ref->xml, "optionmenu_qmmm_eng");
		submenu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));
		item = gtk_menu_get_active(GTK_MENU(submenu));
		
		index = g_list_index(GTK_MENU_SHELL(submenu)->children, item);
		
/*		owner->DiscardCurrentEng();
		
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		// perform a sanity check to give feedback to user, and don't close dialog if failed...
		
		if (owner->CheckEngSettings() == false) return;	*/
		
		delete ref->prj->current_setup;						// get rid of the old setup...
		ref->prj->current_setup = new setup2_qm_mm(ref->prj);			// ...and make a new one.
		
		ref->prj->GetCurrentSetup()->current_eng_index = index;			// set the requested engine.
		break;

#ifdef ENABLE_SIMPLIFIED
		
		case 3:
		delete ref->prj->current_setup;						// get rid of the old setup...
		ref->prj->current_setup = new setup1_sf(ref->prj);			// ...and make a new one.
		break;
		
		case 4:
		delete ref->prj->current_setup;						// get rid of the old setup...
		ref->prj->current_setup = new setup2_mm_sf(ref->prj);			// ...and make a new one.
		break;

#endif	// ENABLE_SIMPLIFIED
		
		default:
		ref->prj->err->ErrorMessage("BUG: Invalid pageinfo at setup_druid::handler_DruidButtonFinish().");
		exit(EXIT_FAILURE);
	}
	
	// report the new settings to log.
	
	char mbuff1[256];
	ostrstream str1(mbuff1, sizeof(mbuff1));
	str1 << "Changed the Setup for calculations ";
	str1 << "(setup = " << ref->prj->GetCurrentSetup()->GetClassName();
	str1 << ", engine = " << ref->prj->GetCurrentSetup()->GetEngineName(ref->prj->GetCurrentSetup()->GetCurrEngIndex());
	str1 << ")." << endl << ends;
	ref->prj->PrintToLog(mbuff1);
	
	// close the window...
	
	gtk_widget_destroy(ref->window);
	
	return TRUE;
}

gboolean setup_druid::handler_DruidButtonCancel(GnomeDruidPage *, gpointer, gpointer data)
{
	setup_druid * ref = (setup_druid *) data;
	cout << "handler_DruidButtonCancel() : ref = " << ref << endl;
	
	// close the window...
	
	gtk_widget_destroy(ref->window);
	return TRUE;
}

/*################################################################################################*/

// eof
