// T3PROJECT.H : the GNOME "project" class.

// Copyright (C) 1998, 2002 Tommi Hassinen.

// This program is free software; you can redistribute it and/or modify it
// under the terms of the license (GNU GPL) which comes with this package.

/*################################################################################################*/

#ifndef T3PROJECT_H
#define T3PROJECT_H

class gnome_project;

/*################################################################################################*/

class graphics_view;		// views.h
class treelist_view;		// views.h
class view_factory;		// views.h

#include "project.h"
#include "t3views.h"

#include "t3dialog.h"

#include <gtk/gtk.h>
#include <gnome.h>

#include <vector>
using namespace std;

#define FILENAME_FILTER	"*.gpr"

/*################################################################################################*/

/**	This will contain GNOME-dependent aspects of the "##project" class.

	A gnome_project object will hold a pointer to GnomeMDIGenericChild; so, it will hold
	the gnome MDI child features.
	
	The another side, the gnome MDI view, is handled as follows: there is only a single
	view, which is a GtkNotebook object. This is a container object that can hold other
	GtkWidgets. So there is only a single MDI view for an MDI child.
*/

class gnome_project : public project
{
	protected:
	
	GnomeMDIGenericChild * child;
	
	GtkWidget * paned_widget;
	GtkWidget * notebook;
	//GtkTextBuffer * log_buffer;	// in GTK2
	GtkWidget * scroll_widget; GtkWidget * log_widget; int logpos;
	
	GtkWidget * popupmenu;
	
	readpdb_mdata * importpdb_mdata;		// temporary?!?!?!
	
	protected:
	
	static GnomeUIInfo filext_submenu[];		// file
	static GnomeUIInfo file_submenu[];
	
	static GnomeUIInfo selmodes_radio[];
	static GnomeUIInfo selmodes_submenu[];
	static GnomeUIInfo select_submenu[];
	
	static GnomeUIInfo settings_submenu[];
	
	static GnomeUIInfo viewscreate_submenu[];	// render/views
	static GnomeUIInfo views_submenu[];		// render
	static GnomeUIInfo lights_submenu[];		// render
	static GnomeUIInfo rendermode_style_radio[];
	static GnomeUIInfo rendermode_color_radio[];
	static GnomeUIInfo rendermode_label_radio[];
	static GnomeUIInfo rendermode_submenu[];	// render
	static GnomeUIInfo projection_radio[];
	static GnomeUIInfo projection_submenu[];	// render
	static GnomeUIInfo render_submenu[];
	
	static GnomeUIInfo objects_submenu[];
	
	static GnomeUIInfo compute_submenu[];
	
	static GnomeUIInfo hydrogens_submenu[];		// build
	static GnomeUIInfo library_submenu[];
	static GnomeUIInfo build_submenu[];
	
	static GnomeUIInfo popupmenu_info[];
	static GnomeUIInfo popupmenu_dummy[];		// for tearoff
	
	friend class gnome_ogl_view;
	friend class gnome_view;
	
	public:
	
	friend class gnome_class_factory;
	friend class gnome_treelist_view;
	friend class gnome_graphics_view;
	
	friend class file_open_dialog;
	friend class file_save_dialog;
	
	friend class importpdb_dialog;		// importpdb_mdata
	
	public:
	
	gnome_project(gnome_class_factory &);
	virtual ~gnome_project(void);
	
	void InitGnomeMDIChild(void);	// BE VERY CAREFUL WITH THIS!!!
	
	static char * GetConfigString(GnomeMDIChild * child, gpointer);
	static GtkWidget * GnomeViewCreator(GnomeMDIChild * child, gpointer);

	void PrintToLog(const char *);			// virtual
	
	void UpdateAllWindowTitles(void);		// virtual
	void event_SelectedObjectChanged(void);		// virtual
	
	void MakeLibraryWidgets(GtkMenuItem *, const char *, int);
	
	// the popup-menu callbacks are here...
	// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	
	static void popup_FileOpen(GtkWidget *, gpointer);
	static void popup_FileImport(GtkWidget *, gpointer);
	static void popup_FileExport(GtkWidget *, gpointer);
	static void popup_FileExportGraphic(GtkWidget *, gpointer);
	static void popup_FileSaveAs(GtkWidget *, gpointer);
	static void popup_FileExtra1(GtkWidget *, gpointer);
	static void popup_FileExtra2(GtkWidget *, gpointer);
	
	static void popup_ViewsNewEnLevDiag(GtkWidget *, gpointer);
	
	static void popup_RModeWireFrame(GtkWidget *, gpointer);
	static void popup_RModeBallAndStick(GtkWidget *, gpointer);
	static void popup_RModeVanDerWaals(GtkWidget *, gpointer);
	static void popup_RModeCylinders(GtkWidget *, gpointer);
	static void popup_RModeNothing(GtkWidget *, gpointer);
	static void popup_CModeState(GtkWidget *, gpointer);
	static void popup_CModeElement(GtkWidget *, gpointer);
	static void popup_LModeNothing(GtkWidget *, gpointer);
	static void popup_LModeIndex(GtkWidget *, gpointer);
	static void popup_LModeCharge(GtkWidget *, gpointer);
	static void popup_LModeElement(GtkWidget *, gpointer);
	static void popup_LModeAtomType(GtkWidget *, gpointer);
	static void popup_LModeBondType(GtkWidget *, gpointer);

	static void popup_ObjRibbon(GtkWidget *, gpointer);
	static void popup_ObjEPlane(GtkWidget *, gpointer);
	static void popup_ObjEVolume(GtkWidget *, gpointer);
	static void popup_ObjESurface(GtkWidget *, gpointer);
	static void popup_ObjVDWSurface(GtkWidget *, gpointer);
	static void popup_ObjEDPlane(GtkWidget *, gpointer);
	static void popup_ObjEDSurface(GtkWidget *, gpointer);
	static void popup_ObjMOPlane(GtkWidget *, gpointer);
	static void popup_ObjMOVolume(GtkWidget *, gpointer);
	static void popup_ObjMOSurface(GtkWidget *, gpointer);
	static void popup_ObjMODPlane(GtkWidget *, gpointer);
	static void popup_ObjMODVolume(GtkWidget *, gpointer);
	static void popup_ObjMODSurface(GtkWidget *, gpointer);
	
	static void popup_CompSetup(GtkWidget *, gpointer);
	static void popup_CompEnergy(GtkWidget *, gpointer);
	static void popup_CompGeomOpt(GtkWidget *, gpointer);
	static void popup_CompMolDyn(GtkWidget *, gpointer);
	static void popup_CompRandomSearch(GtkWidget *, gpointer);
	static void popup_CompSystematicSearch(GtkWidget *, gpointer);
	static void popup_CompMonteCarloSearch(GtkWidget *, gpointer);
	static void popup_CompEPlot1D(GtkWidget *, gpointer);
	static void popup_CompEPlot2D(GtkWidget *, gpointer);
	static void popup_CompFormula(GtkWidget *, gpointer);
	static void popup_TrajView(GtkWidget *, gpointer);
	static void popup_SetOrbital(GtkWidget *, gpointer);
	
	static void popup_HAdd(GtkWidget *, gpointer);
	static void popup_HRemove(GtkWidget *, gpointer);
	static void popup_Clear(GtkWidget *, gpointer);
	static void popup_Library(GtkWidget *, gpointer);
	static void popup_BuilderAmino(GtkWidget *, gpointer);
	static void popup_BuilderNucleic(GtkWidget *, gpointer);
	static void popup_Center(GtkWidget *, gpointer);
	static void popup_EnterCommand(GtkWidget *, gpointer);
	
	void GeomOptGetParam(geomopt_param &);	// virtual
	void MolDynGetParam(moldyn_param &);	// virtual
};

/*################################################################################################*/

#endif	// T3PROJECT_H

// eof
